import einops
import torch
import torch.nn.functional as F
from torch import nn

from kappamodules.init import (
    init_xavier_uniform_zero_bias,
    init_xavier_uniform_merged_linear,
    init_truncnormal_zero_bias,
)


class MMDiTDotProductAttention(nn.Module):
    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=True,
            proj_bias=True,
            seqlens=None,
            channel_first=False,
            init_weights="truncnormal002",
            init_last_proj_zero=False,
    ):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.seqlens = seqlens
        self.channel_first = channel_first
        self.init_weights = init_weights
        self.init_last_proj_zero = init_last_proj_zero

        self.qkv1 = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.qkv2 = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj1 = nn.Linear(dim, dim, bias=proj_bias)
        self.proj2 = nn.Linear(dim, dim, bias=proj_bias)

        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            self.apply(init_xavier_uniform_zero_bias)
            init_xavier_uniform_merged_linear(self.qkv1, num_layers=3)
            init_xavier_uniform_merged_linear(self.qkv2, num_layers=3)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError
        if self.init_last_proj_zero:
            nn.init.zeros_(self.proj.weight)

    @staticmethod
    def to_channel_last(x):
        return einops.rearrange(x, "b c l -> b l c")

    @staticmethod
    def to_channel_first(x):
        return einops.rearrange(x, "b l c -> b c l")

    def forward(self, x1, x2, attn_mask=None):
        if self.channel_first:
            x1 = self.to_channel_last(x1)
            x2 = self.to_channel_last(x2)
        seqlen1 = x1.size(1)
        seqlen2 = x2.size(1)

        q1, k1, v1 = einops.rearrange(
            self.qkv1(x1),
            "bs seqlen (three num_heads head_dim) -> three bs num_heads seqlen head_dim",
            three=3,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)
        q2, k2, v2 = einops.rearrange(
            self.qkv2(x2),
            "bs seqlen (three num_heads head_dim) -> three bs num_heads seqlen head_dim",
            three=3,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)
        # concat in sequence dimension
        q = torch.concat([q1, q2], dim=2)
        k = torch.concat([k1, k2], dim=2)
        v = torch.concat([v1, v2], dim=2)
        # attention
        x = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x1, x2 = x.split([seqlen1, seqlen2], dim=1)
        x1 = self.proj1(x1)
        x2 = self.proj2(x2)

        if self.channel_first:
            x1 = self.to_channel_first(x1)
            x2 = self.to_channel_first(x2)
        return x1, x2



import torch.nn.functional as F
from torch import nn
import torch


class UnetGino(nn.Module):
    def __init__(self, input_dim, hidden_dim, depth=4, num_groups=8, output_dim=None):
        super().__init__()
        self.dim = hidden_dim
        self.depth = depth
        self.num_groups = num_groups
        self.output_dim = output_dim

        # create dim per level
        assert depth > 1
        assert isinstance(hidden_dim, int) and hidden_dim % 2 == 0
        dim_per_level = [hidden_dim * 2 ** i for i in range(depth)]

        # create down path
        self.down_blocks = nn.ModuleList()
        for i in range(depth):
            if i == 0:
                # first block has no pooling and goes from input_dim -> dim_per_level[0] -> dim_per_level[1]
                self.down_blocks.append(
                    nn.Sequential(
                        # pooling
                        nn.Identity(),
                        # conv1
                        nn.GroupNorm(num_groups=1, num_channels=input_dim),
                        nn.Conv3d(input_dim, dim_per_level[0] // 2, kernel_size=3, padding=1, bias=False),
                        nn.ReLU(),
                        # conv2
                        nn.GroupNorm(num_groups=num_groups, num_channels=dim_per_level[0] // 2),
                        nn.Conv3d(dim_per_level[0] // 2, dim_per_level[0], kernel_size=3, padding=1, bias=False),
                        nn.ReLU(),
                    ),
                )
            else:
                self.down_blocks.append(
                    nn.Sequential(
                        # pooling
                        nn.MaxPool3d(kernel_size=2, stride=2),
                        # conv1
                        nn.GroupNorm(num_groups=num_groups, num_channels=dim_per_level[i] // 2),
                        nn.Conv3d(dim_per_level[i] // 2, dim_per_level[i] // 2, kernel_size=3, padding=1, bias=False),
                        nn.ReLU(),
                        # conv2
                        nn.GroupNorm(num_groups=num_groups, num_channels=dim_per_level[i] // 2),
                        nn.Conv3d(dim_per_level[i] // 2, dim_per_level[i], kernel_size=3, padding=1, bias=False),
                        nn.ReLU(),
                    ),
                )

        # create up blocks
        self.up_blocks = nn.ModuleList()
        rev_dim_per_level = list(reversed(dim_per_level))
        for i in range(depth - 1):
            self.up_blocks.append(
                nn.Sequential(
                    # conv1
                    nn.GroupNorm(num_groups=num_groups, num_channels=rev_dim_per_level[i] + rev_dim_per_level[i + 1]),
                    nn.Conv3d(
                        rev_dim_per_level[i] + rev_dim_per_level[i + 1],
                        rev_dim_per_level[i + 1],
                        kernel_size=3,
                        padding=1,
                        bias=False,
                    ),
                    nn.ReLU(),
                    # conv2
                    nn.GroupNorm(num_groups=num_groups, num_channels=rev_dim_per_level[i + 1]),
                    nn.Conv3d(
                        rev_dim_per_level[i + 1],
                        rev_dim_per_level[i + 1],
                        kernel_size=3,
                        padding=1,
                        bias=False,
                    ),
                    nn.ReLU(),
                ),
            )

        self.pred = nn.Conv3d(rev_dim_per_level[-1], output_dim or hidden_dim, kernel_size=1)

    def forward(self, x):
        stack = []

        # down path
        for down_block in self.down_blocks:
            x = down_block(x)
            stack.append(x)

        # x == stack[-1] -> pop last element
        stack.pop()

        # up path
        for up_block in self.up_blocks:
            residual = stack.pop()
            x = F.interpolate(x, scale_factor=2, mode="nearest")
            x = torch.concat([residual, x], dim=1)
            x = up_block(x)

        # pred
        x = self.pred(x)
        return x


import einops
from torch import nn


class LayerNorm1d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.layer = nn.LayerNorm(*args, **kwargs)

    def forward(self, x):
        x = einops.rearrange(x, "batch_size dim height -> batch_size height dim")
        x = self.layer(x)
        x = einops.rearrange(x, "batch_size height dim -> batch_size dim height")
        return x


class LayerNorm2d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.layer = nn.LayerNorm(*args, **kwargs)

    def forward(self, x):
        x = einops.rearrange(x, "batch_size dim height width -> batch_size height width dim")
        x = self.layer(x)
        x = einops.rearrange(x, "batch_size height width dim -> batch_size dim height width")
        return x


class LayerNorm3d(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.layer = nn.LayerNorm(*args, **kwargs)

    def forward(self, x):
        x = einops.rearrange(x, "batch_size dim height width depth -> batch_size height width depth dim")
        x = self.layer(x)
        x = einops.rearrange(x, "batch_size height width depth dim -> batch_size dim height width depth")
        return x



import torch
from torch import nn


class LearnedBatchNorm(nn.Module):
    def __init__(self, dim, affine=True):
        super().__init__()
        self.dim = dim
        self.affine = affine
        self.mean = nn.Parameter(torch.zeros(*self._shape()))
        self.logvar = nn.Parameter(torch.zeros(*self._shape()))
        if affine:
            self.weight = nn.Parameter(torch.ones(*self._shape()))
            self.bias = nn.Parameter(torch.zeros(*self._shape()))
        else:
            self.weight = None
            self.bias = None

    def _shape(self):
        return 1, self.dim

    def forward(self, x):
        # NOTE: cant use F.batch_norm as it is not differentiable to running_mean/running_var
        return (x - self.mean) / self.logvar.exp() * self.weight + self.bias


class LearnedBatchNorm1d(LearnedBatchNorm):
    def _shape(self):
        return 1, self.dim, 1


class LearnedBatchNorm2d(LearnedBatchNorm):
    def _shape(self):
        return 1, self.dim, 1, 1


class LearnedBatchNorm3d(LearnedBatchNorm):
    def _shape(self):
        return 1, self.dim, 1, 1, 1



import einops
import torch
from torch import nn


class ContinuousPosEmbed(nn.Module):
    def __init__(
            self,
            dim: int,
            ndim: int,
            mode: str = "sincos",
            max_value=None,
            interpolation: str = "linear",
            max_wavelength: int = 10000,
            dtype=torch.float32,
    ):
        super().__init__()
        self.dim = dim
        self.ndim = ndim
        self.max_value = max_value
        self.max_wavelength = max_wavelength
        self.mode = mode
        self.interpolation = interpolation

        # if dim is not cleanly divisible -> cut away trailing dimensions
        ndim_padding = dim % ndim
        dim_per_ndim = (dim - ndim_padding) // ndim
        if mode == "sincos":
            assert max_value is None, "max_value is defined, but mode is sincos -> use learnable mode"
            sincos_padding = dim_per_ndim % 2
            self.padding = ndim_padding + sincos_padding * ndim
            eff_dim_per_wave = (self.dim - self.padding) // ndim
            assert eff_dim_per_wave > 0
            self.register_buffer(
                "omega",
                1. / max_wavelength ** (torch.arange(0, eff_dim_per_wave, 2, dtype=dtype) / eff_dim_per_wave),
            )
            self.embed = None
        elif mode == "learnable":
            assert max_value is not None
            if ndim == 1 and not isinstance(max_value, (list, tuple)):
                max_value = [max_value]
            assert len(max_value) == ndim
            self.padding = ndim_padding
            self.embed = nn.ParameterList(
                [
                    nn.Parameter(torch.randn(max_value[i], dim_per_ndim) * 0.02)
                    for i in range(ndim)
                ],
            )
            self.omega = None
        else:
            raise NotImplementedError

    def forward(self, coords):
        if self.mode == "sincos":
            # sine/cosine embedding
            out_dtype = coords.dtype
            ndim = coords.shape[-1]
            assert self.ndim == ndim
            out = coords.unsqueeze(-1).to(self.omega.dtype) @ self.omega.unsqueeze(0)
            emb = torch.concat([torch.sin(out), torch.cos(out)], dim=-1)
            if coords.ndim == 3:
                emb = einops.rearrange(emb, "bs num_points ndim dim -> bs num_points (ndim dim)")
            elif coords.ndim == 2:
                emb = einops.rearrange(emb, "num_points ndim dim -> num_points (ndim dim)")
            else:
                raise NotImplementedError
            emb = emb.to(out_dtype)
        elif self.mode == "learnable":
            if coords.ndim == 1:
                assert self.ndim == 1
                coords = coords.unsqueeze(1)
            assert coords.size(1) == self.ndim
            if coords.dtype == torch.long:
                # no interpolation needed
                emb = torch.concat([self.embed[i][coords[:, i]] for i in range(self.ndim)], dim=1)
            else:
                if self.interpolation == "linear":
                    floored = coords.floor()
                    ceiled = coords.ceil()
                    ceil_weight = coords - floored
                    floor_weight = 1 - ceil_weight
                    emb_floor = torch.concat(
                        [
                            self.embed[i][floored[:, i].long()] * floor_weight[:, i].unsqueeze(1)
                            for i in range(self.ndim)
                        ],
                        dim=1,
                    )
                    emb_ceil = torch.concat(
                        [
                            self.embed[i][ceiled[:, i].long()] * ceil_weight[:, i].unsqueeze(1)
                            for i in range(self.ndim)
                        ],
                        dim=1,
                    )
                    emb = emb_floor + emb_ceil
                else:
                    raise NotImplementedError
        else:
            raise NotImplementedError
        if self.padding > 0:
            padding = torch.zeros(*emb.shape[:-1], self.padding, device=emb.device, dtype=emb.dtype)
            emb = torch.concat([emb, padding], dim=-1)
        return emb

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{type(self).__name__}(dim={self.dim})"



import einops
import numpy as np
from torch import nn

from kappamodules.init import init_truncnormal_zero_bias
from kappamodules.utils.param_checking import to_ntuple
import torch.nn.functional as F
import torch

class VitPatchEmbed(nn.Module):
    def __init__(
            self,
            dim,
            num_channels,
            resolution,
            patch_size,
            stride=None,
            norm_ctor=None,
            flatten=False,
            init_weights="xavier_uniform",
    ):
        super().__init__()
        self.dim = dim
        self.num_channels = num_channels
        self.resolution = resolution
        self.init_weights = init_weights
        self.flatten = flatten
        self.ndim = len(resolution)
        self.patch_size = to_ntuple(patch_size, n=self.ndim)
        if stride is None:
            self.stride = self.patch_size
        else:
            self.stride = to_ntuple(stride, n=self.ndim)
        for i in range(self.ndim):
            assert resolution[i] % self.patch_size[i] == 0, \
                f"resolution[{i}] % patch_size[{i}] != 0 (resolution={resolution} patch_size={patch_size})"
        self.seqlens = [resolution[i] // self.patch_size[i] for i in range(self.ndim)]
        if self.patch_size == self.stride:
            # use primitive type as np.prod gives np.int which is not compatible with all serialization/logging
            self.num_patches = int(np.prod(self.seqlens))
        else:
            if self.ndim == 1:
                conv_func = F.conv1d
            elif self.ndim == 2:
                conv_func = F.conv2d
            elif self.ndim == 3:
                conv_func = F.conv3d
            else:
                raise NotImplementedError
            self.num_patches = conv_func(
                input=torch.zeros(1, 1, *resolution),
                weight=torch.zeros(1, 1, *self.patch_size),
                stride=self.stride,
            ).numel()

        if self.ndim == 1:
            conv_ctor = nn.Conv1d
        elif self.ndim == 2:
            conv_ctor = nn.Conv2d
        elif self.ndim == 3:
            conv_ctor = nn.Conv3d
        else:
            raise NotImplementedError

        self.proj = conv_ctor(num_channels, dim, kernel_size=self.patch_size, stride=self.stride)
        self.norm = nn.Identity() if norm_ctor is None else norm_ctor(dim)
        self.reset_parameters()

    def reset_parameters(self):
        if self.init_weights == "torch":
            pass
        elif self.init_weights == "xavier_uniform":
            # initialize as nn.Linear
            w = self.proj.weight.data
            nn.init.xavier_uniform_(w.view([w.shape[0], -1]))
            nn.init.zeros_(self.proj.bias)
        elif self.init_weights in ["truncnormal", "truncnormal002"]:
            self.apply(init_truncnormal_zero_bias)
        else:
            raise NotImplementedError

    def forward(self, x):
        assert all(x.size(i + 2) % self.patch_size[i] == 0 for i in range(self.ndim)), \
            f"x.shape={x.shape} incompatible with patch_size={self.patch_size}"
        x = self.proj(x)
        if self.flatten:
            x = einops.rearrange(x, "b c ... -> b (...) c")
        else:
            x = einops.rearrange(x, "b c ... -> b ... c")
        x = self.norm(x)
        return x



from .dit_block import DitBlock
from .dit_perceiver_block import DitPerceiverBlock
from .dit_perceiver_pooling_block import DitPerceiverPoolingBlock
from .mlp import Mlp
from .mmdit_block import MMDitBlock
from .perceiver_block import PerceiverBlock
from .perceiver_pooling_block import PerceiverPoolingBlock
from .postnorm_block import PostnormBlock
from .prenorm_block import PrenormBlock



from torch import nn

from kappamodules.init.functional import init_norms_as_noaffine
from .vit_block import VitBlock


class VitBlocks(nn.Module):
    def __init__(self, depth, dim, num_heads, pooling=None, use_last_norm=True, eps=1e-6):
        super().__init__()
        self.depth = depth
        self.dim = dim
        self.num_heads = num_heads
        self.pooling = pooling
        self.use_last_norm = use_last_norm

        self.blocks = nn.Sequential(*[
            VitBlock(
                dim=dim,
                num_heads=num_heads,
                eps=eps,
            )
            for _ in range(depth)
        ])
        if use_last_norm:
            self.norm = nn.LayerNorm(dim, eps=eps)
        else:
            self.norm = nn.Identity()
        self.reset_parameters()

    def reset_parameters(self):
        if self.blocks[-1].init_norms == "torch":
            pass
        elif self.blocks[-1].init_norms == "nonaffine":
            init_norms_as_noaffine(self.norm)
        else:
            raise NotImplementedError

    def forward(self, x):
        x = self.blocks(x)
        if self.pooling is not None:
            x = self.pooling(x)
        x = self.norm(x)
        return x
